"""
Extension for the `RssCog`.
Loading this file via `commands.Bot.load_extension` will add `RssCog` to the bot.
"""

import logging
from typing import Tuple

import validators
from feedparser import FeedParserDict, parse
from discord.ext import commands
from discord import Interaction, Embed, Colour, TextChannel
from discord.app_commands import Choice, Group, autocomplete, choices, rename
from sqlalchemy import insert, select, and_, delete

from utils import get_rss_data, followup, audit  # pylint: disable=E0401
from feed import get_source, Source              # pylint: disable=E0401
from db import (                                 # pylint: disable=E0401
    DatabaseManager,
    SentArticleModel,
    RssSourceModel,
    FeedChannelModel
)

log = logging.getLogger(__name__)

rss_list_sort_choices = [
    Choice(name="Nickname", value=0),
    Choice(name="Date Added", value=1)
]

# TODO SECURITY: a potential attack is that the user submits an rss feed then changes the
# target resource. Run a period task to check this.
async def validate_rss_source(nickname: str, url: str) -> Tuple[str | None, FeedParserDict | None]:
    """Validate a provided RSS source.

    Parameters
    ----------
    nickname : str
        Nickname of the source. Must not contain URL.
    url : str
        URL of the source. Must be URL with valid status code and be an RSS feed.

    Returns
    -------
    str or None
        String invalid message if invalid, NoneType if valid.
    FeedParserDict or None
        The feed parsed from the given URL or None if invalid.
    """

    # Ensure the URL is valid
    if not validators.url(url):
        return f"The URL you have entered is malformed or invalid:\n`{url=}`", None

    # Check the nickname is not a URL
    if validators.url(nickname):
        return "It looks like the nickname you have entered is a URL.\n" \
                f"For security reasons, this is not allowed.\n`{nickname=}`", None


    feed_data, status_code = await get_rss_data(url)

    # Check the URL status code is valid
    if status_code != 200:
        return f"The URL provided returned an invalid status code:\n{url=}, {status_code=}", None

    # Check the contents is actually an RSS feed.
    feed = parse(feed_data)
    if not feed.version:
        return f"The provided URL '{url}' does not seem to be a valid RSS feed.", None

    return None, feed


class RssCog(commands.Cog):
    """
    Command cog.
    """

    def __init__(self, bot):
        super().__init__()
        self.bot = bot

    @commands.Cog.listener()
    async def on_ready(self):
        """Instructions to call when the cog is ready."""

        log.info("%s cog is ready", self.__class__.__name__)

    async def source_autocomplete(self, inter: Interaction, nickname: str):
        """Provides RSS source autocomplete functionality for commands.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        nickname : str
            _description_

        Returns
        -------
        list of app_commands.Choice
            _description_
        """

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == inter.guild_id,
                RssSourceModel.nick.ilike(f"%{nickname}%")
            )
            query = select(RssSourceModel).where(whereclause).order_by(RssSourceModel.nick)
            result = await database.session.execute(query)
            sources = [
                Choice(name=rss.nick, value=rss.rss_url)
                for rss in result.scalars().all()
            ]

        return sources

    # All RSS commands belong to this group.
    feed_group = Group(
        name="feed",
        description="Commands for rss sources.",
        guild_only=True  # We store guild IDs in the database, so guild only = True
    )

    @feed_group.command(name="add")
    async def add_rss_source(self, inter: Interaction, nickname: str, url: str):
        """Add a new RSS source.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        nickname : str
            A name used to identify the RSS source.
        url : str
            The RSS feed URL.
        """

        await inter.response.defer()

        illegal_message, feed = await validate_rss_source(nickname, url)
        if illegal_message:
            await followup(inter, illegal_message, suppress_embeds=True)
            return

        log.debug("RSS feed added")

        async with DatabaseManager() as database:
            query = insert(RssSourceModel).values(
                discord_server_id = inter.guild_id,
                rss_url = url,
                nick=nickname
            )
            await database.session.execute(query)

            await audit(self,
                f"Added RSS source ({nickname=}, {url=})",
                inter.user.id, database=database
            )

        embed = Embed(title="RSS Feed Added", colour=Colour.dark_green())
        embed.add_field(name="Nickname", value=nickname)
        embed.add_field(name="URL", value=url)
        embed.set_thumbnail(url=feed.get("feed", {}).get("image", {}).get("href"))

        await followup(inter, embed=embed)

    @feed_group.command(name="remove")
    @autocomplete(url=source_autocomplete)
    async def remove_rss_source(self, inter: Interaction, url: str):
        """Delete an existing RSS source.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        url : str
            The RSS source to be removed. Autocomplete or enter the URL.
        """

        await inter.response.defer()

        log.debug("Attempting to remove RSS source (url=%s)", url)

        async with DatabaseManager() as database:
            select_result = await database.session.execute(
                select(RssSourceModel).filter(
                    and_(
                        RssSourceModel.discord_server_id == inter.guild_id,
                        RssSourceModel.rss_url == url
                    )
                )
            )
            rss_source = select_result.scalars().one()
            nickname = rss_source.nick

            delete_result = await database.session.execute(
                delete(RssSourceModel).filter(
                    and_(
                        RssSourceModel.discord_server_id == inter.guild_id,
                        RssSourceModel.rss_url == url
                    )
                )
            )

            await audit(self,
                f"Added RSS source ({nickname=}, {url=})",
                inter.user.id, database=database
            )

        if not delete_result.rowcount:
            await followup(inter, "Couldn't find any RSS sources with this name.")
            return

        source = get_source(url)

        embed = Embed(title="RSS Feed Deleted", colour=Colour.dark_red())
        embed.add_field(name="Nickname", value=nickname)
        embed.add_field(name="URL", value=url)
        embed.set_thumbnail(url=source.icon_url)

        await followup(inter, embed=embed)

    @feed_group.command(name="list")
    @choices(sort=rss_list_sort_choices)
    async def list_rss_sources(self, inter: Interaction, sort: Choice[int]=None, sort_reverse: bool=False):
        """Provides a with a list of RSS sources available for the current server.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        # Default to the first choice if not specified.
        if type(sort) is Choice:
            description = "Sort by "
            description += "Nickname " if sort.value == 0 else "Date Added "
            description += '\U000025BC' if sort_reverse else '\U000025B2'
        else:
            sort = rss_list_sort_choices[0]
            description = ""

        sort = sort if type(sort) == Choice else rss_list_sort_choices[0]

        match sort.value, sort_reverse:
            case 0, False:
                order_by = RssSourceModel.nick.asc()
            case 0, True:
                order_by = RssSourceModel.nick.desc()
            case 1, False:                                # NOTE:
                order_by = RssSourceModel.created.desc()  # Datetime order is inversed because we want the latest
            case 1, True:                                 # date first, not the oldest as it would sort otherwise.
                order_by = RssSourceModel.created.asc()
            case _, _:
                raise ValueError("Unknown sort: %s" % sort)

        async with DatabaseManager() as database:
            whereclause = and_(RssSourceModel.discord_server_id == inter.guild_id)
            query = select(RssSourceModel).where(whereclause).order_by(order_by)
            result = await database.session.execute(query)

            rss_sources = result.scalars().all()
            rowcount = len(rss_sources)

            if not rss_sources:
                await followup(inter, "It looks like you have no rss sources.")
                return

            output = "\n".join([f"{i}. **[{rss.nick}]({rss.rss_url})** " for i, rss in enumerate(rss_sources)])

        embed = Embed(
            title="Saved RSS Feeds",
            description=f"{description}\n\n{output}",
            colour=Colour.blue()
        )
        embed.set_footer(text=f"Showing {rowcount} results")

        await followup(inter, embed=embed)

    @feed_group.command(name="fetch")
    @rename(max_="max")
    @autocomplete(rss=source_autocomplete)
    async def fetch_rss(self, inter: Interaction, rss: str, max_: int=1):
        """Fetch an item from the specified RSS feed.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        rss : str
            The RSS feed to fetch from.
        max_ : int, optional
            Maximum number of items to fetch, by default 1, limits at 5.
        """

        await inter.response.defer()

        if max_ > 5:
            followup(inter, "It looks like you have requested too many articles.\nThe limit is 5")
            return

        invalid_message, feed = await validate_rss_source("", rss)
        if invalid_message:
            await followup(inter, invalid_message)
            return

        source = Source.from_parsed(feed)
        articles = source.get_latest_articles(max_)

        if not articles:
            await followup(inter, "Sorry, I couldn't find any articles from this feed.")
            return

        embeds = [await article.to_embed() for article in articles]

        async with DatabaseManager() as database:
            query = insert(SentArticleModel).values([
                {
                    "discord_server_id": inter.guild_id,
                    "discord_channel_id": inter.channel_id,
                    "discord_message_id": inter.id,
                    "article_url": article.url,
                }
                for article in articles
            ])
            await database.session.execute(query)
            await audit(self, f"User is requesting {max_} articles from {source.name}", inter.user.id, database=database)

        await followup(inter, embeds=embeds)


    # Channels ---- ---- ----


    async def autocomplete_rss_sources(self, inter: Interaction, nickname: str):
        """"""

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == inter.guild_id,
                RssSourceModel.nick.ilike(f"%{nickname}%")
            )
            query = select(RssSourceModel).where(whereclause)
            result = await database.session.execute(query)
            sources = [
                Choice(name=rss.nick, value=rss.id)
                for rss in result.scalars().all()
            ]

        log.debug("Autocomplete rss_sources returned %s results", len(sources))

        return sources

    async def autocomplete_existing_feeds(self, inter: Interaction, current: str):
        """Returns a list of existing RSS + Channel feeds.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        current : str
            The current text entered for the autocomplete.
        """

        async with DatabaseManager() as database:
            whereclause = and_(
                FeedChannelModel.discord_server_id == inter.guild_id,
                FeedChannelModel.search_name.ilike(f"%{current}%"),  # is this secure from SQL Injection atk ?
                RssSourceModel.id == FeedChannelModel.rss_source_id
            )
            query = (
                select(FeedChannelModel, RssSourceModel)
                .where(whereclause)
                .join(RssSourceModel)
                .order_by(FeedChannelModel.discord_channel_id)
            )
            result = await database.session.execute(query)
            feeds = []
            for feed in result.scalars().all():
                channel = inter.guild.get_channel(feed.discord_channel_id)
                feeds.append(Choice(name=f"# {channel.name}  |  {feed.rss_source.nick}", value=feed.id))

        log.debug("Autocomplete existing_feeds returned %s results", len(feeds))

        return feeds

    # # All RSS commands belong to this group.
    # channel_group = Group(
    #     name="channels",
    #     description="Commands for channel assignment.",
    #     guild_only=True  # These commands belong to channels of
    # )

    @feed_group.command(name="assign")
    @autocomplete(rss=autocomplete_rss_sources)
    async def include_feed(self, inter: Interaction, rss: int, channel: TextChannel = None):
        """Include a feed within the specified channel.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        rss : str
            The RSS feed to include.
        channel : TextChannel
            The channel to include the feed in.
        """

        await inter.response.defer()

        channel = channel or inter.channel

        async with DatabaseManager() as database:
            select_query = select(RssSourceModel).where(and_(
                RssSourceModel.id == rss,
                RssSourceModel.discord_server_id == inter.guild_id
            ))

            select_result = await database.session.execute(select_query)
            rss_source = select_result.scalars().one()
            nick, rss_url = rss_source.nick,  rss_source.rss_url

            insert_query = insert(FeedChannelModel).values(
                discord_server_id = inter.guild_id,
                discord_channel_id = channel.id,
                rss_source_id=rss,
                search_name=f"{nick} #{channel.name}"
            )

            await database.session.execute(insert_query)


        await followup(inter, f"I've included [{nick}]({rss_url}) to {channel.mention}")

    @feed_group.command(name="unassign")
    @autocomplete(option=autocomplete_existing_feeds)
    async def exclude_feed(self, inter: Interaction, option: int):
        """Undo command for the `/channel include-feed` command.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        option : str
            The RSS feed and channel to exclude.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            query = delete(FeedChannelModel).where(and_(
                FeedChannelModel.id == option,
                FeedChannelModel.discord_server_id == inter.guild_id
            ))

            result = await database.session.execute(query)

        if not result.rowcount:
            await followup(inter, "I couldn't find any items under that ID (placeholder response)")
            return

        await followup(inter, "I've removed this item (placeholder response)")

    @feed_group.command(name="channels")
    # @choices(sort=[
    #     Choice(name="RSS Nickname", value=0),
    #     Choice(name="Channel ID", value=1),
    #     Choice(name="Date Added", value=2)
    # ])
    async def list_feeds(self, inter: Interaction):  # sort: int
        """List all of the channels and their respective included feeds.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            whereclause = and_(
                FeedChannelModel.discord_server_id == inter.guild_id,
                RssSourceModel.id == FeedChannelModel.rss_source_id
            )
            query = (
                select(FeedChannelModel, RssSourceModel)
                .where(whereclause)
                .join(RssSourceModel)
                .order_by(FeedChannelModel.discord_channel_id)
            )
            result = await database.session.execute(query)

            feed_channels = result.scalars().all()
            rowcount = len(feed_channels)

            if not feed_channels:
                await followup(inter, "It looks like there are no feed channels available.")
                return

            output = "\n".join([
                f"{i}. <#{feed.discord_channel_id}> · [{feed.rss_source.nick}]({feed.rss_source.rss_url})"
                for i, feed in enumerate(feed_channels)
            ])

        embed = Embed(
            title="Saved Feed Channels",
            description=f"{output}",
            colour=Colour.blue()
        )
        embed.set_footer(text=f"Showing {rowcount} results")

        await followup(inter, embed=embed)



async def setup(bot):
    """
    Setup function for this extension.
    Adds `RssCog` to the bot.
    """

    cog = RssCog(bot)
    await bot.add_cog(cog)
    log.info("Added %s cog", cog.__class__.__name__)
